#!/usr/bin/python3
from configparser import NoOptionError
from configparser import ConfigParser as SafeConfigParser
import os.path
import os
import shutil
import logging
import threading
import subprocess
from SystemUpdater.Core.utils import UpdaterKconf2Err
import json
from typing import Dict,Any
from pathlib import Path
from copy import deepcopy

class UpgradeConfig(SafeConfigParser):
    # 备份还原之前保存的配置
    UPDATE_NOT_RESOTRE_DIR = "/var/lib/kylin-system-updater/remain"

    def __init__(self, datadir="/var/lib/kylin-system-updater/", name="system-updater.conf",defaults_conf=None):
        SafeConfigParser.__init__(self)
        self.datadir = datadir
        self.maincfg = os.path.join(datadir, name)

        self.config_files = []
        # 增加的默认配置，先读取默认配置加载配置项，最终写入到主配置内
        if defaults_conf:
            self.config_files.append(os.path.join(datadir, defaults_conf))
        self.config_files.append(self.maincfg)
        
        # 检查是否存在备份的配置，存在时直接移动到默认位置
        self.remail_conf = os.path.join(self.UPDATE_NOT_RESOTRE_DIR, name)
        if os.path.exists(self.remail_conf):
            shutil.move(self.remail_conf,self.maincfg)

        # 检查配置文件是否存在
        for conf in self.config_files:
            if not os.path.exists(conf):
                logging.warning("Check: Conf(%s) not exits...",conf)
                # self.config_files.remove(conf)

        self.read(self.config_files, encoding='iso-8859-1')
        logging.info("Initialize Upgrade ConfigFile(%s) to success",str(self.config_files))

    def reRead(self):
        self.read(self.config_files, encoding='iso-8859-1')

    def setValue(self, section, option, value=None,is_write = True):
        logging.info("SetValue Section:%s Option:%s Value:%s",section, option, value)
        try:
            self.reRead()
            self.set(section, option,value)
        except Exception as e:
                logging.error("Error: setValue section:%s option:%s value:%s",section, option, value)
                logging.error(str(e))
                return False
        if is_write == True:
            with open(self.maincfg, 'w+') as configfile:
                self.write(configfile)
        return True

    def getWithDefault(self, section, option, default,re_read=False):
        try:
            if re_read == True:
                self.reRead()

            if type(default) == bool:
                return self.getboolean(section, option)
            elif type(default) == float:
                return self.getfloat(section, option)
            elif type(default) == int:
                return self.getint(section, option)
            return self.get(section, option)
        except Exception as e:
            logging.warning(str(e))
            return default

    def getlist(self, section, option):
        try:
            tmp = self.get(section, option)
        except Exception as e:
            logging.error(str(e))
            return []
        items = [x.strip() for x in tmp.split(" ")]
        if '' in items and len(items):
            return []
        return items

    def setListValue(self, section, option, value=None,is_write = True):
        tmp = str(value).replace('[', '').replace(']', '')
        tmp = tmp.replace("'", '').replace(',', '')
        try:
            self.set(section, option,tmp)
        except Exception as e:
            logging.error("setListValue section:%s option:%s",section, option)
            logging.error(str(e))
            return
        if is_write == True:
            with open(self.maincfg, 'w+') as configfile:
                self.write(configfile)

    def getListFromFile(self, section, option):
        try:
            filename = self.get(section, option)
        except NoOptionError:
            return []
        p = os.path.join(self.datadir, filename)
        if not os.path.exists(p):
            logging.error("getListFromFile: no '%s' found" % p)
        with open(p) as f:
            items = [x.strip() for x in f]
        return [s for s in items if not s.startswith("#") and not s == ""]

    # 还原之前保留配置
    def remain_conf(self):
        shutil.copy(self.maincfg, self.remail_conf)

    def optionxform(self, optionstr: str) -> str:
        return optionstr

class UpgradeKConf2Manager():
    APPLICATION_NAME = 'kylin-update-frontend'
    GROUP_TO_LISTEN  = [
        'kylin-update-frontend.notification',
        'kylin-update-frontend.downloadlimit',
        'kylin-update-frontend.autodownload'
        ]

    def __init__(self, dbusCtrl = None, skip_listen = False):
        self.dbusControlor  = dbusCtrl
        self.pty_lock       = threading.Lock()
        self.master_fds     = []
        self.should_exit    = False

        if not skip_listen:
            self.conf2_listen_keys_changed()
    
    def quitKConf2Manager(self):
        self.should_exit = True

    def _parse_kconf2_output(self, output):
        updater_kconf2_childrens = []

        if "Wrong schema id" != str(output).strip():
            updater_kconf2_childrens = output.splitlines()

        return updater_kconf2_childrens

    def _parse_kconf2_value(self, output):
        return str(output).strip()
    
    def _get_updater_kconf2_childrens(self):
        updater_kconf2_childrens = None

        args = ["kconf2", "list-children", self.APPLICATION_NAME]

        process = subprocess.Popen(" ".join(args), shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        output, error = process.communicate()
        if not error:
            updater_kconf2_childrens = self._parse_kconf2_output(output.decode('utf-8'))
        else:
            logging.error("UpgradeConfigKConf2: %s",str(error))

        return updater_kconf2_childrens

    def conf2_listen_keys_changed(self):
        try:
            if not os.path.exists("/usr/bin/kconf2"):
                logging.error("conf2_listen_keys_changed: kconf2 not found.")
                return

            updater_kconf2_childrens = self._get_updater_kconf2_childrens()
            for ikc in updater_kconf2_childrens:
                if ikc in self.GROUP_TO_LISTEN:
                    monitor_thread = threading.Thread(target=self._listen_key_changed, args=(ikc,))
                    monitor_thread.start()

        except Exception as e:
            logging.error("UpgradeConfigKConf2: %s",str(e))

    def _listen_key_changed(self, schema):
        try:
            import pty
            import select

            master_fd = -1
            slave_fd  = -1

            master_fd, slave_fd = pty.openpty()

            args = ["kconf2", "monitor", schema]
            process = subprocess.Popen(args, stdout=slave_fd, stderr=slave_fd, text=True)
            # logging.warning("kconf2 monitor [%s %d]", schema, threading.get_ident())

            os.close(slave_fd)

            self.master_fds.append(master_fd)

            while not self.should_exit:
                rlist, _, _ = select.select([master_fd], [], [], 1)

                if master_fd in rlist:
                    output = os.read(master_fd, 1024).decode('utf-8')
                    if output:
                        with self.pty_lock:
                            # logging.warning("[%s %d]:%s", schema, threading.get_ident(), output.strip())

                            change_details = [part.strip() for part in output.strip().split('changed to')]

                            if schema == 'kylin-update-frontend.notification':

                                if "enable" == str(change_details[0]):
                                    self.dbusControlor.AutocheckStatusChanged(
                                                        True if str(change_details[-1]).lower() == 'true' else False)

                            elif schema == 'kylin-update-frontend.downloadlimit':

                                if "enable" == str(change_details[0]):
                                    downloadlimitValue = self.conf2_get_key_value(schema, 'value')
                                    self.dbusControlor.DownloadStatusChanged(
                                                        True if str(change_details[-1]).lower() == 'true' else False,
                                                        int(downloadlimitValue))
                                
                                elif "value" == str(change_details[0]):
                                    downloadspeedLimitStatus = self.conf2_get_key_value(schema, 'enable')
                                    self.dbusControlor.DownloadStatusChanged(
                                                        True if downloadspeedLimitStatus.lower() == 'true' else False,
                                                        int(str(change_details[-1])) if str(change_details[-1]).isdigit() else int(0))

                            elif schema == 'kylin-update-frontend.autodownload':

                                if "enable" == str(change_details[0]):
                                    self.dbusControlor.AutoUpgradeStatusChanged(
                                                        True if str(change_details[-1]).lower() == 'true' else False)

            process.terminate()
        except Exception as e:
            logging.error("UpgradeConfigKConf2: %s",str(e))

    def conf2_get_key_value(self, schema, key):
        updater_kconf2_value = ''
        
        if not os.path.exists("/usr/bin/kconf2"):
            raise UpdaterKconf2Err()

        try:
            args = ["kconf2", "get", schema, key]

            process = subprocess.Popen(" ".join(args), shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            output, error = process.communicate()
            if not error:
                updater_kconf2_value = self._parse_kconf2_value(output.decode('utf-8'))
            else:
                logging.warning(str(error))
            logging.info("UpgradeConfigKConf2: %s:%s -> %s.", schema, key, updater_kconf2_value)
        except Exception as e:
            logging.error("UpgradeConfigKConf2: %s",str(e))

        return updater_kconf2_value

    def conf2_set_key_value(self, schema, key, value):
        setStatus = True
        try:
            args = ["kconf2", "set", schema, key, value]

            process = subprocess.Popen(" ".join(args), shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            output, error = process.communicate()
            logging.info("UpgradeConfigKConf2: %s:%s -> %s.", schema, key, value)
        except Exception as e:
            logging.error("UpgradeConfigKConf2: %s",str(e))
            setStatus = False

        return setStatus

class UpgradeJsonManager():
    def __init__(self,filelist=["/var/lib/unattended-upgrades/kylin-update-strategy.json"]):
        initial_data=self.merge_files(filelist)
        self._data = deepcopy(initial_data) if initial_data else {}

    def reload_config(self,filelist=["/usr/share/kylin-system-updater/kylin-update-strategy.json"]):
        initial_data=self.merge_files(filelist)
        self._data = deepcopy(initial_data) if initial_data else {}
        
    def merge_files(self,file_list: list) -> dict:
        """
        合并多个JSON文件内容，遇到异常时返回空字典

        参数:
            file_list: 需要合并的文件路径列表

        返回:
            dict: 合并后的字典数据，异常时返回{}
        """
        merged_data = {}

        for file_path in file_list:
            try:
                # 检查文件存在性
                if not os.path.exists(file_path):
                    raise FileNotFoundError(f"文件 {file_path} 不存在")

                # 检查文件可读性
                if not os.access(file_path, os.R_OK):
                    raise PermissionError(f"无权限读取文件 {file_path}")

                # 读取并解析文件
                with open(file_path, 'r', encoding='utf-8') as f:
                    data = json.load(f)
                    merged_data.update(data)

            except (FileNotFoundError, PermissionError) as e:
                (f"[错误] {str(e)}")
                return {}
            except json.JSONDecodeError:
                logging.error(f"[错误] 文件 {file_path} 不是有效的JSON格式")
                return {}
            except Exception as e:
                logging.error(f"[未知错误] 处理文件 {file_path} 时发生意外错误: {str(e)}")
                return {}

        return merged_data
    
        # 直接返回当前JSON对象（深拷贝安全模式）
    
    def get_json(self, indent: int = None) -> str:
        """
        输出内部JSON对象的字符串形式
        
        参数:
            indent (int): 缩进空格数，None表示紧凑格式
            
        返回:
            str: JSON格式字符串
        """
        return json.dumps(
            self.data,  # 使用@property的深拷贝
            indent=indent,
            ensure_ascii=False,
            default=str  # 处理无法序列化的对象
        )
    
    def get_json_object(self) -> dict:
        """
        返回内部JSON对象的深拷贝副本
        
        返回:
            dict: 原始数据结构，包含当前所有键值对
        """
        return self.data  # 复用已存在的安全机制
    
    def update_and_diff(self, new_data: Dict[str, Any]) -> Dict[str, Dict]:
        """
        更新数据并输出差异
        返回结构示例：
        {
            "added": {"new_key": "value"},
            "modified": {"existing_key": {"old": "value", "new": "updated"}}
        }
        """
        diff_result = {"added": {}, "modified": {}}
        self._recursive_merge(self._data, new_data, diff_result)
        return diff_result
    
    def _recursive_merge(self, internal: Dict, external: Dict, diff: Dict, path: str = "") -> None:
        """深度递归合并并记录差异"""
        for key, ext_value in external.items():
            current_path = f"{path}.{key}" if path else key
            
            # 处理新增键
            if key not in internal:
                internal[key] = deepcopy(ext_value)
                diff["added"][current_path] = ext_value
                continue
                
            int_value = internal[key]
            # 处理嵌套字典
            if isinstance(int_value, dict) and isinstance(ext_value, dict):
                self._recursive_merge(int_value, ext_value, diff, current_path)
            # 处理值差异
            elif int_value != ext_value:
                diff["modified"][current_path] = {
                    "old": deepcopy(int_value),
                    "new": deepcopy(ext_value)
                }
                internal[key] = deepcopy(ext_value)

    def save_to_file(self, file_path: str = "/var/lib/unattended-upgrades/kylin-update-strategy.json", indent: int = None, ensure_ascii: bool = False) -> bool:
        """
        将内部JSON对象保存到指定文件
        
        参数:
            file_path: 目标文件路径（支持绝对/相对路径）
            indent: 缩进空格数（None为紧凑格式）
            ensure_ascii: 是否转义非ASCII字符）
            
        返回:
            bool: 保存是否成功
        """
        try:
            # 创建目录
            directory = os.path.dirname(file_path)
            if directory and not os.path.exists(directory):
                os.makedirs(directory, exist_ok=True)
                
            # 深拷贝后写入
            with open(file_path, 'w', encoding='utf-8') as f:
                json.dump(
                    self.data,  # 使用@property的深拷贝
                    f,
                    indent=indent,
                    ensure_ascii=ensure_ascii,
                    separators=(',', ': '),
                    default=str  # 处理不可序列化对象
                )
            return True
        except (IOError, PermissionError) as e:
            logging.error(f"文件保存失败: {str(e)}")
            return False
        except Exception as e:
            logging.error(f"未知错误: {str(e)}")
            return False

    @property
    def data(self) -> Dict[str, Any]:
        """获取数据的深拷贝副本"""
        return deepcopy(self._data)

    def __str__(self) -> str:
        """美化输出"""
        return json.dumps(self._data, indent=2, ensure_ascii=False)
    


if __name__ == "__main__":
    pass